"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2021 Deepkit UG, Marc J. Schmidt
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the MIT License.
 *
 * You should have received a copy of the MIT License along with this program.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PostgresDatabaseAdapter = exports.PostgresSQLDatabaseQueryFactory = exports.PostgresSQLDatabaseQuery = exports.PostgresSQLQueryResolver = exports.PostgresPersistence = exports.PostgresConnectionPool = exports.PostgresDatabaseTransaction = exports.PostgresConnection = exports.PostgresStatement = void 0;
const sql_1 = require("@deepkit/sql");
const orm_1 = require("@deepkit/orm");
const postgres_platform_js_1 = require("./postgres-platform.js");
const pg_1 = __importDefault(require("pg"));
const core_1 = require("@deepkit/core");
const stopwatch_1 = require("@deepkit/stopwatch");
const type_1 = require("@deepkit/type");
const config_js_1 = require("./config.js");
/**
 * Converts a specific database error to a more specific error, if possible.
 */
function handleSpecificError(session, error) {
    let cause = error;
    while (cause) {
        if (cause instanceof Error) {
            if (cause.message.includes('duplicate key value')
                && 'table' in cause && 'string' === typeof cause.table
                && 'detail' in cause && 'string' === typeof cause.detail) {
                return new orm_1.UniqueConstraintFailure(`${cause.message}: ${cause.detail}`, { cause: error });
            }
            cause = cause.cause;
        }
    }
    return error;
}
class PostgresStatement extends sql_1.SQLStatement {
    constructor(logger, sql, client, stopwatch) {
        super();
        this.logger = logger;
        this.sql = sql;
        this.client = client;
        this.stopwatch = stopwatch;
        this.released = false;
    }
    async get(params = []) {
        const frame = this.stopwatch ? this.stopwatch.start('Query', stopwatch_1.FrameCategory.databaseQuery) : undefined;
        try {
            if (frame)
                frame.data({ sql: this.sql, sqlParams: params });
            this.logger.logQuery(this.sql, params);
            //postgres driver does not maintain error.stack when they throw errors, so
            //we have to manually convert it using asyncOperation.
            const res = await (0, core_1.asyncOperation)((resolve, reject) => {
                this.client.query(this.sql, params).then(resolve).catch(reject);
            });
            return res.rows[0];
        }
        catch (error) {
            error = (0, orm_1.ensureDatabaseError)(error);
            this.logger.failedQuery(error, this.sql, params);
            throw error;
        }
        finally {
            if (frame)
                frame.end();
        }
    }
    async all(params = []) {
        const frame = this.stopwatch ? this.stopwatch.start('Query', stopwatch_1.FrameCategory.databaseQuery) : undefined;
        try {
            if (frame)
                frame.data({ sql: this.sql, sqlParams: params });
            this.logger.logQuery(this.sql, params);
            //postgres driver does not maintain error.stack when they throw errors, so
            //we have to manually convert it using asyncOperation.
            const res = await (0, core_1.asyncOperation)((resolve, reject) => {
                this.client.query(this.sql, params).then(resolve).catch(reject);
            });
            return res.rows;
        }
        catch (error) {
            error = (0, orm_1.ensureDatabaseError)(error);
            this.logger.failedQuery(error, this.sql, params);
            throw error;
        }
        finally {
            if (frame)
                frame.end();
        }
    }
    release() {
    }
}
exports.PostgresStatement = PostgresStatement;
class PostgresConnection extends sql_1.SQLConnection {
    constructor(connectionPool, connection, logger, transaction, stopwatch) {
        super(connectionPool, logger, transaction, stopwatch);
        this.connection = connection;
        this.changes = 0;
        this.lastReturningRows = [];
    }
    async prepare(sql) {
        return new PostgresStatement(this.logger, sql, this.connection, this.stopwatch);
    }
    async run(sql, params = []) {
        const frame = this.stopwatch ? this.stopwatch.start('Query', stopwatch_1.FrameCategory.databaseQuery) : undefined;
        try {
            if (frame)
                frame.data({ sql, sqlParams: params });
            //postgres driver does not maintain error.stack when they throw errors, so
            //we have to manually convert it using asyncOperation.
            const res = await (0, core_1.asyncOperation)((resolve, reject) => {
                this.connection.query(sql, params).then(resolve).catch(reject);
            });
            this.logger.logQuery(sql, params);
            this.lastReturningRows = res.rows;
            this.changes = res.rowCount;
        }
        catch (error) {
            error = (0, orm_1.ensureDatabaseError)(error);
            this.logger.failedQuery(error, sql, params);
            throw error;
        }
        finally {
            if (frame)
                frame.end();
        }
    }
    async getChanges() {
        return this.changes;
    }
}
exports.PostgresConnection = PostgresConnection;
class PostgresDatabaseTransaction extends orm_1.DatabaseTransaction {
    /**
     * This is the default for mysql databases.
     */
    repeatableRead() {
        this.setTransaction = 'REPEATABLE READ';
        return this;
    }
    readCommitted() {
        this.setTransaction = 'READ COMMITTED';
        return this;
    }
    serializable() {
        this.setTransaction = 'SERIALIZABLE';
        return this;
    }
    async begin() {
        if (!this.connection)
            return;
        const set = this.setTransaction ? 'SET TRANSACTION ISOLATION LEVEL ' + this.setTransaction + ';' : '';
        await this.connection.run(set + 'START TRANSACTION');
    }
    async commit() {
        if (!this.connection)
            return;
        if (this.ended)
            throw new Error('Transaction ended already');
        await this.connection.run('COMMIT');
        this.ended = true;
        this.connection.release();
    }
    async rollback() {
        if (!this.connection)
            return;
        if (this.ended)
            throw new Error('Transaction ended already');
        await this.connection.run('ROLLBACK');
        this.ended = true;
        this.connection.release();
    }
}
exports.PostgresDatabaseTransaction = PostgresDatabaseTransaction;
class PostgresConnectionPool extends sql_1.SQLConnectionPool {
    constructor(pool) {
        super();
        this.pool = pool;
    }
    async getConnection(logger, transaction, stopwatch) {
        //when a transaction object is given, it means we make the connection sticky exclusively to that transaction
        //and only release the connection when the transaction is commit/rollback is executed.
        if (transaction && transaction.connection)
            return transaction.connection;
        const poolClient = await this.pool.connect();
        this.activeConnections++;
        const connection = new PostgresConnection(this, poolClient, logger, transaction, stopwatch);
        if (transaction) {
            transaction.connection = connection;
            try {
                await transaction.begin();
            }
            catch (error) {
                transaction.ended = true;
                connection.release();
                throw new Error('Could not start transaction: ' + error);
            }
        }
        return connection;
    }
    release(connection) {
        //connections attached to a transaction are not automatically released.
        //only with commit/rollback actions
        if (connection.transaction && !connection.transaction.ended)
            return;
        super.release(connection);
        connection.connection.release();
    }
}
exports.PostgresConnectionPool = PostgresConnectionPool;
function typeSafeDefaultValue(property) {
    if (property.type.kind === type_1.ReflectionKind.string)
        return '';
    if (property.type.kind === type_1.ReflectionKind.number)
        return 0;
    if (property.type.kind === type_1.ReflectionKind.boolean)
        return false;
    if (property.type.kind === type_1.ReflectionKind.class && property.type.classType === Date)
        return false;
    return null;
}
class PostgresPersistence extends sql_1.SQLPersistence {
    constructor(platform, connectionPool, session) {
        super(platform, connectionPool, session);
        this.platform = platform;
        this.connectionPool = connectionPool;
    }
    handleSpecificError(error) {
        return handleSpecificError(this.session, error);
    }
    async batchUpdate(entity, changeSets) {
        const prepared = (0, sql_1.prepareBatchUpdate)(this.platform, entity, changeSets);
        if (!prepared)
            return;
        const placeholderStrategy = new this.platform.placeholderStrategy();
        const params = [];
        const selects = [];
        const valuesValues = [];
        const valuesSetValues = [];
        const valuesNames = [];
        const valuesSetNames = [];
        for (const fieldName of prepared.changedFields) {
            valuesNames.push(fieldName);
            valuesSetNames.push('_changed_' + fieldName);
        }
        for (let i = 0; i < changeSets.length; i++) {
            params.push(prepared.primaryKeys[i]);
            let pkValue = entity.primaryKey.sqlTypeCast(placeholderStrategy.getPlaceholder());
            valuesValues.push('(' + pkValue + ',' + prepared.changedProperties.map(property => {
                params.push(prepared.values[property.name][i]);
                return property.sqlTypeCast(placeholderStrategy.getPlaceholder());
            }).join(',') + ')');
        }
        for (let i = 0; i < changeSets.length; i++) {
            params.push(prepared.primaryKeys[i]);
            let valuesSetValueSql = entity.primaryKey.sqlTypeCast(placeholderStrategy.getPlaceholder());
            for (const fieldName of prepared.changedFields) {
                valuesSetValueSql += ', ' + prepared.valuesSet[fieldName][i];
            }
            valuesSetValues.push('(' + valuesSetValueSql + ')');
        }
        for (const i of prepared.changedFields) {
            const col = entity.fieldMap[i].columnNameEscaped;
            const colChanged = '_changed_' + i;
            if (prepared.aggregateSelects[i]) {
                const select = [];
                select.push('CASE');
                for (const item of prepared.aggregateSelects[i]) {
                    select.push(`WHEN _.${prepared.originPkField} = ${item.id} THEN ${item.sql}`);
                }
                select.push(`ELSE (CASE WHEN ${colChanged} = 0 THEN _origin.${col} ELSE _.${col} END) END as ${col}`);
                selects.push(select.join(' '));
            }
            else {
                //if(check, true, false) => COALESCE(NULLIF(check, true), false)
                selects.push(`(CASE WHEN ${colChanged} = 0 THEN _origin.${col} ELSE _.${col} END) as ${col}`);
            }
        }
        const returningSelect = [];
        returningSelect.push(prepared.tableName + '.' + prepared.pkField);
        if (!(0, core_1.empty)(prepared.setReturning)) {
            for (const i in prepared.setReturning) {
                returningSelect.push(prepared.tableName + '.' + this.platform.quoteIdentifier(i));
            }
        }
        const escapedValuesNames = valuesNames.map(v => entity.fieldMap[v].columnNameEscaped);
        const sql = `
              WITH _b(${prepared.originPkField}, ${escapedValuesNames.join(', ')}) AS (
                SELECT _.${prepared.originPkField}, ${selects.join(', ')} FROM
                    (VALUES ${valuesValues.join(', ')}) as _(${prepared.originPkField}, ${escapedValuesNames.join(', ')})
                    INNER JOIN (VALUES ${valuesSetValues.join(', ')}) as _set(${prepared.pkField}, ${valuesSetNames.join(', ')}) ON (_.${prepared.originPkField} = _set.${prepared.pkField})
                    INNER JOIN ${prepared.tableName} as _origin ON (_origin.${prepared.pkField} = _.${prepared.originPkField})
              )
              UPDATE ${prepared.tableName}
              SET ${prepared.setNames.join(', ')}
              FROM _b
              WHERE ${prepared.tableName}.${prepared.pkField} = _b.${prepared.originPkField}
              RETURNING ${returningSelect.join(', ')};
        `;
        try {
            const connection = await this.getConnection(); //will automatically be released in SQLPersistence
            const result = await connection.execAndReturnAll(sql, params);
            for (const returning of result) {
                const r = prepared.assignReturning[returning[prepared.pkName]];
                if (!r)
                    continue;
                for (const name of r.names) {
                    r.item[name] = returning[name];
                }
            }
        }
        catch (error) {
            const reflection = type_1.ReflectionClass.from(entity.type);
            error = new orm_1.DatabaseUpdateError(reflection, changeSets, `Could not update ${reflection.getClassName()} in database`, { cause: error });
            throw this.handleSpecificError(error);
        }
    }
    async populateAutoIncrementFields(classSchema, items) {
        const autoIncrement = classSchema.getAutoIncrement();
        if (!autoIncrement)
            return;
        const connection = await this.getConnection(); //will automatically be released in SQLPersistence
        //We adjusted the INSERT SQL with additional RETURNING which returns all generated
        //auto-increment values. We read the result and simply assign the value.
        const name = autoIncrement.name;
        const insertedRows = connection.lastReturningRows;
        if (!insertedRows.length)
            return;
        for (let i = 0; i < items.length; i++) {
            items[i][name] = insertedRows[i][name];
        }
    }
    getInsertSQL(classSchema, fields, values) {
        const autoIncrement = classSchema.getAutoIncrement();
        const returning = autoIncrement ? ` RETURNING ${this.platform.quoteIdentifier(autoIncrement.name)}` : '';
        if (fields.length === 0) {
            const pkName = this.platform.quoteIdentifier(classSchema.getPrimary().name);
            fields.push(pkName);
            values.fill('DEFAULT');
        }
        return `INSERT INTO ${this.platform.getTableIdentifier(classSchema)} (${fields.join(', ')}) VALUES (${values.join('), (')}) ${returning}`;
    }
}
exports.PostgresPersistence = PostgresPersistence;
class PostgresSQLQueryResolver extends sql_1.SQLQueryResolver {
    async delete(model, deleteResult) {
        const primaryKey = this.classSchema.getPrimary();
        const pkField = this.platform.quoteIdentifier(primaryKey.name);
        const primaryKeyConverted = (0, orm_1.primaryKeyObjectConverter)(this.classSchema, this.platform.serializer.deserializeRegistry);
        const sqlBuilder = new sql_1.SqlBuilder(this.adapter);
        const tableName = this.platform.getTableIdentifier(this.classSchema);
        const select = sqlBuilder.select(this.classSchema, model, { select: [`${tableName}.${pkField}`] });
        const connection = await this.connectionPool.getConnection(this.session.logger, this.session.assignedTransaction, this.session.stopwatch);
        try {
            const sql = `
                WITH _ AS (${select.sql})
                DELETE
                FROM ${tableName} USING _
                WHERE ${tableName}.${pkField} = _.${pkField}
                RETURNING ${tableName}.${pkField}
            `;
            const rows = await connection.execAndReturnAll(sql, select.params);
            deleteResult.modified = rows.length;
            for (const row of rows) {
                deleteResult.primaryKeys.push(primaryKeyConverted(row[primaryKey.name]));
            }
        }
        catch (error) {
            error = new orm_1.DatabaseDeleteError(this.classSchema, 'Could not delete in database', { cause: error });
            error.query = model;
            throw this.handleSpecificError(error);
        }
        finally {
            connection.release();
        }
    }
    handleSpecificError(error) {
        return handleSpecificError(this.session, error);
    }
    async patch(model, changes, patchResult) {
        const select = [];
        const selectParams = [];
        const entity = (0, sql_1.getPreparedEntity)(this.session.adapter, this.classSchema);
        const tableName = entity.tableNameEscaped;
        const primaryKey = this.classSchema.getPrimary();
        const primaryKeyConverted = (0, orm_1.primaryKeyObjectConverter)(this.classSchema, this.platform.serializer.deserializeRegistry);
        const fieldsSet = {};
        const aggregateFields = {};
        const patchSerialize = (0, type_1.getPatchSerializeFunction)(this.classSchema.type, this.platform.serializer.serializeRegistry);
        const $set = changes.$set ? patchSerialize(changes.$set, undefined) : undefined;
        const set = [];
        if ($set)
            for (const i in $set) {
                if (!$set.hasOwnProperty(i))
                    continue;
                if ($set[i] === undefined || $set[i] === null) {
                    set.push(`${this.platform.quoteIdentifier(i)} = NULL`);
                }
                else {
                    fieldsSet[i] = 1;
                    select.push(`$${selectParams.length + 1} as ${this.platform.quoteIdentifier((0, sql_1.asAliasName)(i))}`);
                    selectParams.push($set[i]);
                }
            }
        if (changes.$unset)
            for (const i in changes.$unset) {
                if (!changes.$unset.hasOwnProperty(i))
                    continue;
                fieldsSet[i] = 1;
                select.push(`NULL as ${this.platform.quoteIdentifier(i)}`);
            }
        for (const i of model.returning) {
            aggregateFields[i] = { converted: (0, type_1.getSerializeFunction)((0, type_1.resolvePath)(i, this.classSchema.type), this.platform.serializer.deserializeRegistry) };
            select.push(`(${this.platform.quoteIdentifier(i)} ) as ${this.platform.quoteIdentifier(i)}`);
        }
        if (changes.$inc)
            for (const i in changes.$inc) {
                if (!changes.$inc.hasOwnProperty(i))
                    continue;
                fieldsSet[i] = 1;
                aggregateFields[i] = { converted: (0, type_1.getSerializeFunction)((0, type_1.resolvePath)(i, this.classSchema.type), this.platform.serializer.serializeRegistry) };
                const sqlTypeCast = (0, sql_1.getDeepTypeCaster)(entity, i);
                select.push(`(${sqlTypeCast('(' + this.platform.getColumnAccessor('', i) + ')')} + ${this.platform.quoteValue(changes.$inc[i])}) as ${this.platform.quoteIdentifier((0, sql_1.asAliasName)(i))}`);
            }
        for (const i in fieldsSet) {
            if (i.includes('.')) {
                let [firstPart, secondPart] = (0, sql_1.splitDotPath)(i);
                const path = '{' + secondPart.replace(/\./g, ',').replace(/[\]\[]/g, '') + '}';
                set.push(`${this.platform.quoteIdentifier(firstPart)} = jsonb_set(${this.platform.quoteIdentifier(firstPart)}, '${path}', to_jsonb(_b.${this.platform.quoteIdentifier((0, sql_1.asAliasName)(i))}))`);
            }
            else {
                const property = entity.fieldMap[i];
                const ref = '_b.' + this.platform.quoteIdentifier((0, sql_1.asAliasName)(i));
                set.push(`${this.platform.quoteIdentifier(i)} = ${property.sqlTypeCast(ref)}`);
            }
        }
        let bPrimaryKey = primaryKey.name;
        //we need a different name because primaryKeys could be updated as well
        if (fieldsSet[primaryKey.name]) {
            select.unshift(this.platform.quoteIdentifier(primaryKey.name) + ' as __' + primaryKey.name);
            bPrimaryKey = '__' + primaryKey.name;
        }
        else {
            select.unshift(this.platform.quoteIdentifier(primaryKey.name));
        }
        const returningSelect = [];
        returningSelect.push(tableName + '.' + this.platform.quoteIdentifier(primaryKey.name));
        if (!(0, core_1.empty)(aggregateFields)) {
            for (const i in aggregateFields) {
                returningSelect.push(this.platform.getColumnAccessor(tableName, i));
            }
        }
        const sqlBuilder = new sql_1.SqlBuilder(this.adapter, selectParams);
        const selectSQL = sqlBuilder.select(this.classSchema, model, { select });
        const sql = `
            WITH _b AS (${selectSQL.sql})
            UPDATE
                ${tableName}
            SET ${set.join(', ')}
            FROM _b
            WHERE ${tableName}.${this.platform.quoteIdentifier(primaryKey.name)} = _b.${this.platform.quoteIdentifier(bPrimaryKey)}
                RETURNING ${returningSelect.join(', ')}
        `;
        const connection = await this.connectionPool.getConnection(this.session.logger, this.session.assignedTransaction, this.session.stopwatch);
        try {
            const result = await connection.execAndReturnAll(sql, selectSQL.params);
            patchResult.modified = result.length;
            for (const i in aggregateFields) {
                patchResult.returning[i] = [];
            }
            for (const returning of result) {
                patchResult.primaryKeys.push(primaryKeyConverted(returning[primaryKey.name]));
                for (const i in aggregateFields) {
                    patchResult.returning[i].push(aggregateFields[i].converted(returning[i]));
                }
            }
        }
        catch (error) {
            error = new orm_1.DatabasePatchError(this.classSchema, model, changes, `Could not patch ${this.classSchema.getClassName()} in database`, { cause: error });
            throw this.handleSpecificError(error);
        }
        finally {
            connection.release();
        }
    }
}
exports.PostgresSQLQueryResolver = PostgresSQLQueryResolver;
class PostgresSQLDatabaseQuery extends sql_1.SQLDatabaseQuery {
}
exports.PostgresSQLDatabaseQuery = PostgresSQLDatabaseQuery;
class PostgresSQLDatabaseQueryFactory extends sql_1.SQLDatabaseQueryFactory {
    createQuery(type) {
        return new PostgresSQLDatabaseQuery(type_1.ReflectionClass.from(type), this.databaseSession, new PostgresSQLQueryResolver(this.connectionPool, this.platform, type_1.ReflectionClass.from(type), this.databaseSession.adapter, this.databaseSession));
    }
}
exports.PostgresSQLDatabaseQueryFactory = PostgresSQLDatabaseQueryFactory;
class PostgresDatabaseAdapter extends sql_1.SQLDatabaseAdapter {
    constructor(options, additional = {}) {
        super();
        this.platform = new postgres_platform_js_1.PostgresPlatform();
        this.closed = false;
        const defaults = {};
        options = 'string' === typeof options ? (0, config_js_1.parseConnectionString)(options) : options;
        this.options = Object.assign(defaults, options, additional);
        this.pool = new pg_1.default.Pool(this.options);
        this.connectionPool = new PostgresConnectionPool(this.pool);
        pg_1.default.types.setTypeParser(1700, parseFloat);
        pg_1.default.types.setTypeParser(20, parseInt);
    }
    getName() {
        return 'postgres';
    }
    getSchemaName() {
        //todo extract schema name from connection options. This acts as default when a table has no schemaName defined.
        return '';
    }
    createPersistence(session) {
        return new PostgresPersistence(this.platform, this.connectionPool, session);
    }
    createTransaction(session) {
        return new PostgresDatabaseTransaction;
    }
    queryFactory(session) {
        return new PostgresSQLDatabaseQueryFactory(this.connectionPool, this.platform, session);
    }
    disconnect(force) {
        if (this.closed)
            return;
        this.closed = true;
        this.pool.end().catch(console.error);
    }
}
exports.PostgresDatabaseAdapter = PostgresDatabaseAdapter;
//# sourceMappingURL=postgres-adapter.js.map